import functools
import json
from json import JSONDecodeError
from typing import Callable, Dict, List
from . import loghelp
from . import errors
from . errors import ProgrammerError, ProgrammerErrorHandle, SafeCatchAll


class EventError(errors.Error):
    """The base class for all errors in this module
    """
    ...


class EventSubscriberCallbackError:
    """Class of error parenting all errors related to subscriber callbacks
    """
    ...


class EventSubscriberCallbackErrorNotCallable(EventSubscriberCallbackError):
    """Error raised when a subscriber callback is not callable
    """
    ...


class Event():
    """Event class represents an event that can be posted and subscribed to

    Attributes:
      channel: str
        The channel to which the event is said to be existing in
      name: str
        The name of the event, used for easy identification
      payload: str
        The payload of the event, used to convey the information of the event, usually in JSON format
    """
    channel: str = ...
    name: str = ...

    payload: str = ...

    def __init__(self, *, channel: str = ..., name: str = ..., payload: str = ..., **kwargs) -> None:
        self.channel = channel
        if self.channel is ...:
            # TODO add warning for instinating event without channel handle
            ...
        if type(self.channel) is not str:
            # TODO add warning for instinating event with non-serializable (not str) channel handle
            ...
        self.name = name
        if self.name is ...:
            # TODO add warning for instinating event without name handle
            ...
        if type(self.name) is not str:
            # TODO add warning for instinating event with non-serializable (not str) name handle
            ...
        self.payload = payload
        if self.payload is ...:
            # TODO add warning for instinating event without payload
            ...
        if type(self.payload) is not str:
            # TODO add warning for instinating event with non-serializable (not str) payload
            ...

    def send(self, /, stage=..., **kwargs) -> None:
        """Posts the event to the given EventStage as if EventStage.post(event=self) was called (hint it is : )

          Args:       
            stage: EventStage
              The stage to which the event is sent too
              Note: Annotation for stage is not possible as this is a convenience method
        """
        try:
            stage.post(event=self)
        except SafeCatchAll as err:
            # TODO add warning for eventstage post failure
            raise err


class HeartBeatEvent(Event):
    """Event class representing a heartbeat event

    For __docs__ on Event, see Event.__doc__ NOT here!
    Attributes:
      channel: str = "Smg88::HeartBeat"
        The channel to which the event is said to be existing in
      name: str = f"Smg88 HeartBeat ({self.count}) at about {loghelp.now()}"
      payload: str = JSON Format below:
        {
          "count": int,
          "approxtime": str,
        }
    """
    count: int = ...

    def __init__(self, /, count: int = -1, *, channel: str = ..., name: str = ..., timestr: str = ..., payload: str = ..., **kwargs) -> None:
        self.count = count
        if type(self.count) is not int:
            # TODO add warning for non-serializable (not int) count
            ...
        self.timestr = timestr
        if self.timestr is ...:
            self.timestr = loghelp.now()
        if type(self.timestr) is not str:
            # TODO add warning for non-serializable (not str) timestr
            ...
        self.payload = payload
        if self.payload is ...:
            self.payload = {
                "count": self.count,
                "approxtime": self.timestr,
            }
        self.name = name
        if self.name is ...:
            self.name = f"Smg88 HeartBeat ({self.count}) at about {self.timestr}"
        if type(self.name) is not str:
            # TODO add warning for non-serializable (not str) name
            ...
        try:
            self._package = json.dumps(self.payload)
        except SafeCatchAll as err:
          # TODO properly handle this error :)
            ...
        super().__init__(channel=channel, name=name, payload=payload, **kwargs)


class EventStage():
    """Represents a place for events to occur
    Events are objects that have a channel and a name, all events are passed to all subscribers of that event's channel

    Events can be posted to this stage by instancing the 'Event' object and calling EventStageInstance.post(EventInstance)
    Or, shorthand, call 'send' on an Event instance with the EventStage instance

    Methods:
      subscribe(callback)
        Subscribes the function to be called with all Event objects posted to this stage under the channel (default __name__ of function)

    Attributes:
      nameHandle: str (constructable)
        A common name for the EventStage, needed to connect to the EventStage
      subscriptions: Dict[str, Callable.__name__]
        Is a property exposing the subscriptions of the stage, the key is the channel and the value is the name of the subscriber's function

      _subscriptions: Dict[str, Callable]
        A dictionary of channel names to functions that are subscribed to that channel name
      _eventBuffer: List[Event]
        A buffer of events to post to the stage
    """
    _subscriptions: Dict[str, List[Callable]]
    nameHandle: str

    _eventBuffer: List[Event]

    @property
    def eventBuffer(self) -> List[Event]:
        return self._eventBuffer

    @property
    def channels(self) -> List[str]:
        return list(self._subscriptions.keys())

    def __init__(self, /, nameHandle: str = ...) -> None:
        self.nameHandle = nameHandle
        if self.nameHandle is ...:
            # TODO add warning for instinating an EventStage without a nameHandle
            ...
        if type(self.nameHandle) is not str:
            # TODO add warning for instinating an EventStage without a serializable (str) nameHandle
            ...
        self._subscriptions = {}
        self._eventBuffer = []

    def post(self, /, event: Event = ...,) -> None:
        if event is ...:
            # TODO add warning for not passing an event
            raise errors.InappropriateRequest(f"No event was passed to the post method {event}",
                                              errorHandle=errors.ProgrammerErrorHandle("Must pass an event to the post method (of an EventStage instance or child of such)"))
        self._eventBuffer.append(event)

    def release(self, /, channel: str = ...,) -> None:
        """_posts all events in the buffer that are in the given channel

        Overloads:
          release()
            Release the latest event in the buffer: _post(1, all=False, retain=False)
          release("channel")
            Release all of the events conforming to that channel: _release(channel="channel")

        Args:
            channel (str, optional): Channel to post all buffered events that are in.
        """
        if channel is ...:
            self._post(num=1, all=False, retain=False)
        else:
            self._release(channel=channel)

    def _release(self, /, channel: str = ...,) -> None:
        if channel is ...:
            raise errors.InappropriateRequest("No channel was passed to the _release method", errorHandle=errors.ProgrammerErrorHandle(
                "Must pass a channel to the _release method (of an EventStage instance or child of such)"))
        [self._handle(event, remove=True)
         for event in self._eventBuffer if event.channel == channel]

    def subscribe(self, /, callback: Callable = ..., *, channel: str = ...) -> None:
        """Subscribes a callback to the given channel (defaults to callback.__name__)

        Args:
            callback (Callable): Callable to call with event=event when event is posted on that channel. Defaults to ....
            channel (str, optional): The exact channel to subscribe the callback to. Defaults to callback.__name__
        """
        # TODO add some info for this function as it is very useful
        if channel is ...:
            channel = callback.__name__
        print(f"Subscribing to EventStage {callback=} under {channel=}")
        if channel not in self.channels:
            self._subscriptions[channel] = []
        self._subscriptions[channel].append(callback)

    def _post(self, /, num: int = 1, *, all: bool = False, retain: bool = ..., **kwargs) -> None:
        if all:
            if retain is ...:
                retain = True
            if not retain:
                # TODO Warn for purging buffer
                ...
            self._postn(num=len(self._eventBuffer),
                        retain=bool(retain), **kwargs)
        else:
            self._postn(num=num, retain=retain, **kwargs)

    def _postn(self, /, num: int = 1, *, retain: bool = False, **kwargs) -> None:
        for _ in range(num):
            self._handle(self._eventBuffer.pop(0))

    def _handle(self, /, event: Event, *, remove: bool = ...) -> None:
        if remove is ...:
            remove = False
        if event.channel in self.channels:
            subscribers = self._subscriptions[event.channel]
            for subscriber in subscribers:
                # TODO add warning for subscriber callback error
                subscriber(event=event)
        else:
            # TODO add warning for no subscribers to given event channel
            ...
        if remove:
            self._eventBuffer.remove(event)


class EventStageHeartbeat():
    """Represents a heartbeat for an AutoEventStage, subscribes to its own channel and reposts it with count++

    Attributes:
      SubscribeHandle: Callable
        Represents the handle 

    Methods:
      pump(): Pump this heartbeat once more!
      subscribe
    """

    counter: int = ...
    approxlastpump: str = ...

    stages: List[EventStage] = ...
    defaultChannel: str = "Smg88::heartbeat::pulse"

    @staticmethod
    def __subscribeHandle(event: Event = ...) -> None:
        """Internal function to be called on a heartbeat

        Args:
            event (Event): Event to be analysed
        """
        print("Cool event handle called!")
        print(f"{event=}")

    def __init__(self, *, stage: EventStage = ..., stages: List[EventStage] = ..., countstart: int = -1) -> None:
        self.counter = countstart
        if type(self.counter) is not int:
            # TODO add warning for non-int counter
            raise errors.InappropriateRequest("counterstart must be an int", errorHandle=ProgrammerErrorHandle(
                "counterstart must be an int when instinating an EventStageHeartbeat object (or children of such)"))
        self.stages = []
        if stage is ... and stages is ...:
            # TODO add info for not passing a stage or stages
            ...
        if stage is not ... and stages is not ...:
            raise errors.InappropriateRequest("Cannot pass both a stage and stages", errorHandle=ProgrammerErrorHandle(
                "Please pass only a stages or a list of stages to an EventStageHeartbeat object constructor (or children thereof)"))
        if stage is ...:
            # TODO add info for not passing a stage
            ...
        else:
            self.stages.append(stage)
        if stages is ...:
            # TODO add info for not passing stages
            ...
        else:
            for stage in stages:
                self.stages.append(stage)

    def pump(self) -> None:
        self._step()

    def _step(self) -> None:
        self.counter += 1
        self.postdefault()
        [stage.release(channel=self.defaultChannels) for stage in self.stages]

    def _defaultEventConstructor(self):
        return HeartBeatEvent(count=self.counter,)

    def postdefault(self) -> None:
        [stage.post(event=self._defaultEventConstructor())
         for stage in self.stages]

    def _subscribeTo(self, *, stage: EventStage = ..., channel: str = ...) -> None:
        """Internal function to subscribe this heartbeat to an EventStage

        Args:
            stage (EventStage): Stage to subscribe to
        """
        if channel is ...:
            # TODO add info for calling _subscribeTo with no given channel
            channel = self.defaultChannel
        if stage is ...:
            # TODO add error for calling _subscribeTo without a given stage
            raise errors.InappropriateRequest(
                "stage not given to _subscribeTo")

        @stage.subscribe
        @loghelp.callbacknamed(channel)
        def _(event: Event = ...):
            self.__subscribeHandle(event)

    def setupStage(self, /, stage: EventStage = ...) -> None:
        """Setups up a stage to receive heartbeats from this object

        Args:
            stage (EventStage): Stage to setup (NOT a list of stages)
        """
        if stage is ...:
            # TODO add warning for calling setup without a stage
            raise errors.InappropriateRequest("stage not given to setupStage")
        self._subscribeTo(stage=stage, channel=self.defaultChannel)


class AutoEventStage(EventStage):
    """An EventStage that automatically posts its events (no manual post required)

    Args:
        nameHandle: str
        autosetup: bool
        heartbeat: EventStageHeartbeat
          Dependency injection only
    """
    heartbeat: EventStageHeartbeat = ...

    def setup(self, *, heartbeat: EventStageHeartbeat = ...) -> None:
        if heartbeat is ...:
            self.heartbeat = EventStageHeartbeat()
        if type(self.heartbeat) is not EventStageHeartbeat:
            # TODO add warning for non-EventStageHeartbeat heartbeat
            ...
        self.heartbeat.setupStage(stage=self)

        self.heartbeat.pump()

    def __init__(self, *, nameHandle: str = ..., autosetup: bool = ..., heartbeat: EventStageHeartbeat = ...) -> None:
        super().__init__(nameHandle=nameHandle)
        self.heartbeat = heartbeat
        if self.heartbeat is ...:
            self.heartbeat = EventStageHeartbeat(stage=self, countstart=0)
        if autosetup:
            self.setup(heartbeat=self.heartbeat)


def main():
    stage = AutoEventStage()

    @stage.subscribe
    @loghelp.callbacknamed("Smg88")
    def _(event: Event):
        print(
            f"EVENT {event=}, {event.channel=}, {event.name=}, {event.payload=}")
    stage.post(Event(channel="Smg", name="help!", payload="TESTING!"))
    stage.post(Event(channel="Smg88", name="LETS F**KING GO!", payload="gout!"))
    stage._post(all=True)


if __name__ == "__main__":
    main()



from hashlib import sha256
import sys
import string

"""
Keys
Note: most up to date record is kept in Discord::Discord Raiders Hub::For Private Reference::Improved Alts

appleid --str
smartguy89 --str > gmail + discord
smartguy88 --str > discord + gmail
smartguy87 --str > gmail + discord
smartguy86 --str > gmail + discord
"""

"""
Password that use --str

apple
smartguy89
smartguy88
smartguy87
smartguy86
"""

chars = string.ascii_letters + string.digits + string.punctuation + \
    " " + string.punctuation + string.ascii_letters + \
    string.digits + string.punctuation
if len(sys.argv) > 4:
  print(f"Chars ({len(chars)}):{chars}")


def main(args):
  global chars
  try:
    input = bytes(args[0], encoding="utf-8")
    extra = bytes(args[1] or b"", encoding="utf-8")
  except:
    print("Invalid input, requires 2 text command line inputs")
  else:
    testObj = sha256(input + extra)
    testHash = str(testObj.hexdigest())
    print(f"Full Hash:{testHash}")
    print(f"Standard Length Hash:{testHash[:10]}")
    testHashNum = str(int.from_bytes(testObj.digest(), "big"))
    # Splits string every 2nd character
    testHashNumList = [testHashNum[i:i+2]
                       for i in range(0, len(testHashNum), 2)]
    #print(testHashStr) # Debug only :)
    hashStr = "".join([chars[int(testHashNumList[n]) + int(testHashNumList[n+1])]
                      for n in range(0, len(testHashNumList)-1, 2)])
    finalHashStr = hashStr[:5] if len(hashStr) > 5 else hashStr
    if len(args) > 2:
      print(f"Hash String:Smg!88{finalHashStr}")
  print("Finish")


if __name__ == "__main__":
  main(sys.argv[1:])



from enum import auto
from typing import List, overload
from Smg88.errors import ProgrammerErrorHandle
from loghelp import EnumParent
import errors
from errors import Error


class PointRole(EnumParent):
    Server = auto()
    Node = auto()


class PointType(EnumParent):
    LocalToLocal = auto()
    LocalToRemote = auto()
    RemoteToLocal = auto()
    RemoteToRemote = auto()


class PacketItem():
    _content: List[str]

    def __init__(self, *strings):
        self._content = strings

    @property
    def content(self):
        return self._content

    @property
    def msg(self):
        return self._content[0]

    def __str__(self) -> str:
        return self.msg


class Communicator():
    """Represents a communication 'stream' or option for communicating with something, usually a Point

    isasync: bool = True
    requestRole(): PointRole
    requestType(): PointType
    buffer: List[Item] property get
    bufferSend: List[Item] property get

    get(num: int): Item
        Takes (num number of) items off the _buffer and gives it to the caller
    send(*items: Item): Item
        Takes items and puts it on the _bufferSend

    _buffer: List[Item]
        An array that represents the data received  
    _bufferSend: List[Item]
        An array that represents data yet to be sent (on next eventloop perhaps?)

    _send(num: int): Item
        Asynchronously sends (num number of) items from the _bufferSend start
    _receive(*packets): Item
        Is called to receive packet/s, appends then to _buffer end

    """
    isasync: bool

    _buffer: List[str]
    _bufferSend: List[PacketItem]

    def __init__(self, isasync: bool = True):
        self.isasync = isasync
        self._buffer = []
        self._bufferSend = []

    @property
    def buffer(self):
        return self._buffer

    @property
    def bufferSend(self):
        return self._bufferSend

    def get(self, num: int):
        raise NotImplementedError

    def send(self, *packets: PacketItem):
        raise NotImplementedError

    def _receive(self, *packets: PacketItem):
        raise NotImplementedError

    def _send(self, *packets: PacketItem):
        raise NotImplementedError


class CommunicatorErrors(errors.Errors):
    """Errors for the Communicator class"""
    ...


class CommunicatorBufferError(CommunicatorErrors):
    """Errors for the Communicator class related to buffers"""
    ...


class CommunicatorEmptyBuffer(CommunicatorBufferError):
    """Raised when strict and the buffer is empty"""
    ...


class CommunicatorPurgeBuffer(CommunicatorBufferError):
    """Raised when strict and the buffer is purged"""
    ...


class CLICommunicator(Communicator):
    def send(self, *packets: PacketItem) -> None:
        (self._bufferSend.append(packet) for packet in packets)

    @overload
    def get(self, num: int = 1) -> List[PacketItem]:
        ...

    @overload
    def get(self, all: bool = True) -> List[PacketItem]:
        ...

    def get(self, /, num: int = 1, *, all: bool = False, retain: bool = ..., strict: bool = True) -> List[PacketItem]:
        """Gets (num number of) items from the buffer

        Args:
            num (int, optional): how many items to take from the buffer. Defaults to 1.
            all (bool, optional *overrides num*): whether to take all items from the buffer. Defaults to False.
            retain (bool, optional): whether to retain taken items or to pop them. Defaults to False, unless all then *defaults* to True
            strict (bool, optional): whether to raise errors on bad practices. Defaults to True.

        Raises:
            CommunicatorPurgeBuffer: Raises when strict and the buffer is purged, as in all=True and retain=False (note: retain defaults to True when all=True to avoid this)
            CommunicatorEmptyBuffer: Raises when the items requested (through setting num) are not in the buffer

        Returns:
            List[PacketItem]: A list of PacketItems taken from the buffer
        """
        if all:
            num = len(self._buffer)
            if not retain:
                # TODO Add warning for purging self._buffer when !strict
                if strict:
                    raise CommunicatorPurgeBuffer(errorHandle=ProgrammerErrorHandle(
                        "Try not to set all=True and retain=False else you will purge (delete the whole of) the self._buffer!"))
        elif retain is ...:
            retain = False
        if type(num) is not int:
            raise errors.InappropriateRequest("num must be an int if provided and all=False", errorHandle=ProgrammerErrorHandle(
                "Either set all=True or provide an int for the num parameter when using get()"))
        if num > len(self._buffer):
            raise CommunicatorEmptyBuffer(errorHandle=ProgrammerErrorHandle(
                "Do not request more items than are in the buffer, use all=True and/or retain=True"))
        if retain:
            return [self._buffer[i] for i in range(num)]
        return [self._buffer.pop(0) for i in range(num)]

    def _receive(self, *packets: PacketItem) -> None:
        self._bufffer.append(packets)

    def _send(self, num: int = 1) -> None:
        print(f"P{i}: {self._buffer.pop(0)}\n" for i in range(num))


class Communication():
    """Represents a possible communication 'target' and provides an array of Communicator objects to use to communicate with is

    communicators: List[Communicator]
    requestRole(): PointRole
    requestType(): PointType
    """

    def __init__(self) -> None:
        ...


class Point():
    """Class that represents a single point (to current python interpreter) in the IOT

    role: PointRole
    type: PointType

    requestPoints(): Dict
        Returns a dictionary of Communication objects for each point found that responded in my IOT
    requestServers(): Dict
        Returns a dictionary of Communication objects for each server found that responded in my IOT
    requestNodes(): Dict
        Returns a dictionary of Communication objects for each node found that responded in my IOT

    """

    def __init__(self):
        ...


def tests():
    """Tests for the coms.py module"""
    p = Point()
    test = CLICommunicator()
    test.send(PacketItem("Hello world!"))
    test.send(test.get())
    ...


if __name__ == "__main__":
    tests()



"""This module adds supportive functions and classes for logging to the Smg88 package
Supports Smg88.events, Smg88.coms, Smg88.security



Should Include:
Syntactic sugar for logging
"""

from datetime import datetime
from enum import Enum, auto, unique
import logging
from typing import Callable

from . import errors

# Logging levels are as follows:
# Note that commented definitions are default for the logging module anyway

# logging.NOTSET = 0
logging.RAW_DEBUG = 3
logging.USER_RAW_DEBUG = 4
# logging.DEBUG = 9
logging.STANDARD_DEBUG = 13
logging.HELPFUL_DEBUG = 14
# logging.INFO = 19
logging.STANDARD_INFO = 23
logging.HELPFUL_INFO = 24
# logging.WARNING = 29
logging.STANDARD_WARNING = 33
logging.HELPFUL_WARNING = 34
# logging.ERROR = 39
logging.STANDARD_ERROR = 43
logging.HELPFUL_ERROR = 44
logging.PRIVATE_ERROR = 48
# logging.CRITICAL = 49
logging.STANDARD_CRITICAL = 53
logging.HELPFUL_CRITICAL = 54
logging.PRIVATE_CRITICAL = 57
logging.FATAL_CRITICAL = 58
logging.CORRUPTED_CRITICAL = 59

# Which channels should I use?
# logging.NOTSET # for None (sentinel value)
logging.USER_RAW_DEBUG  # for raw debugging
logging.HELPFUL_DEBUG  # for debugging generally
logging.HELPFUL_INFO  # for logging info
logging.HELPFUL_WARNING  # for logging warnings
# for logging errors, specifically recoverable errors by reloading or some other specified method
logging.HELPFUL_ERROR
logging.HELPFUL_CRITICAL  # for logging critical, specifically fatal, error

# Complete non-user-friendly documentation for logging levels:
"""
**An actual description of logging levels are as follows:**
  # NOTSET = 0, this is a sentinel value representing a level that is not defined (not set)
  > RAW_DEBUG = 3, this is the lowest level of debug messages used internally by the bot
  > USER_RAW_DEBUG = 4, this is the lowest level of debug messages used by the user and is the encouraged lowest level of debug message for the user
  
  # DEBUG = 9, this is not a sentinel value but is the original value of debug for the logging module
  > STANDARD_DEBUG = 13, this is the normal level of debug used by the bot and will typically describe internal events
  > HELPFUL_DEBUG = 14, this is the recommended user debug level for most things just debugging
  
  # INFO = 19, this is not a sentinel value but is the original value of info for the logging module
  > STANDARD_INFO = 23, this is the normal level of info used by the bot and will typically describe internal actions with a meaningful timestamp
  > HELPFUL_INFO = 24, this is the recommended user info level for most things just informational, try to aim for something meaningful maybe with a timestamp?
  
  # WARNING = 29, this is not a sentinel value but is the original value of warning for the logging module
  > STANDARD_WARNING = 33, this is the normal level of warnings by the bot and will typically only include warnings of significance that inhibit actual functionality, implementation warnings are typically found in STANDARD_INFO
  > HELPFUL_WARNING = 34, this is the recommended user warning level for most warnings, although you can really use any arbitrary number logs >30 will be treated with slightly more 'respect' and are designed around being informative and meaningful

  # ERROR = 39, this is not a sentinel value but is the original value of error for the logging module
  > STANDARD_ERROR = 43, this is the normal level of error used by the bot and will typically describe internal events that are recoverably *only* with a restart / reloading process, concerning internal / implementation details
  > HELPFUL_ERROR = 44, this is the recommended user error level for most errors, although you can really use any arbitrary number logs >40 will be treated with slightly more 'oomph' and are designed around being informative, meaningful, and with a workaround or reload process to circumvent the issue / error. If this is not possible, i.e. recovery from the error, than put it into a *_CRITICAL logging level such as 50

  # CRITICAL = 49, this is not a sentinel value but is the original value of critical for the logging module
  > STANDARD_CRITICAL = 53, this is the normal level of critical used by the bot and will typically describe internal events that are irrecoverably period. This is the level that should be used when the bot is unable to recover from the error, or when the bot is unable to continue functioning.
  > HELPFUL_CRITICAL = 54, this is the recommended user critical level for most errors, although you can really use any arbitrary number logs >50 will be treated with slightly more 'oomph' and are designed around being informative, meaningful, but sadly un-recoverable. Restarting / reloading is probably not going to solve the problem
  > PRIVATE_CRITICAL = 57, this is a **private** logging level used to typically DM (if still possible) an admin (or myself) of the error and a basic error handle / code. Rigging this up is *NOT* an added feature (at time of writing) and used only as a failsafe
  > FATAL_CRITICAL = 58, this is a generic logging level used to indicate a, well, fatally critical error of any kind ONLY related to *the programming side*, so not something like a loss of internet or battery failure e.t.c.
  > CORRUPTED_CRITICAL = 59, this is a logging level dedicated to uncontrollable, fatal corruption of programming, or some other generic corruption. If a SEU (Single Event Upset) occurs, this is the level that should most of the time be 'raised' on :)

**Logging levels num summary:**
  1-4 = Raw debug
  4-10 = User Raw debug

  # CUTOFF for private data, all logs >9 should not contain any private data such as the discord bots token or any other sensitive data
  10-14 = Standard debug
  14-19 = Helpful debug

  19-24 = Standard info
  24-29 = Helpful info

  29-34 = Standard warning
  34-39 = Helpful warning

  39-44 = Standard error
  44-49 = Helpful error

  # CUTOFF for implementation-detail data, all logs >=39 should contain __user friendly__ text, so that a general user could understand the problem at least generally. Try not to include details related to the actual implementation of the bot
  50-54 = Standard critical
  54-57 = Helpful critical

  57-58 = Private critical # Think a personal DM message
  58-59 = Fatal critical
  59-60 = Corrupted critical # Think an impossible result such as corrupted data or a SEU
"""


class log():
    def __init__(self, msg, *, level=0):
        self.msg = msg


class EnumParent(Enum):
    def _generate_next_value_(name, start, count, last_values):
        return f"{name}"


def now():
    return datetime.now().strftime("%d/%m/%Y %H:%M:%S")


def callbacknamed(name: str = ...):
    if name is ...:
        # TODO add warning for using decorator without given name
        errors.InappropriateRequest(
            "Name not given to callbacknamed decorator constructor")

    def __decoratorfunction(func: Callable = ...):
        if func is ...:
            # TODO add warning for using func decorator without a given function ??
            raise errors.InappropriateRequest(
                "WTF? Decorator used without given function?")
        func.__name__ = name
        return func
    return __decoratorfunction



from coms import Communicator
comm = Communicator()
import time
import discord

intents = discord.Intents.default()
intents.message_content = True

client = discord.Client(intents=intents)
TOKEN = comm.requestSecret("ryanpinger TOKEN")


@client.event
async def on_ready():
    print(f'We have logged in as {client.user}')

_MYID = 634212810275946496  # My id
_RYANID = 312130522283245570  # Ryans id

DebugChannel = None
TargetID = _RYANID
TargetMember = None
TargetChannel = None
STOPPINGING = False


@client.event
async def on_message(message):
    global DebugChannel
    global TargetChannel
    global TargetMember
    global TargetID
    global STOPPINGING
    if message.content.startswith("$$START"):
        STOPPINGING = False
        await message.channel.send(f"Started pinging :) you can now live in fear :thumbsup:")
        await DebugChannel.send(f"Started pinging :) :thumbsup:")
    if message.content.startswith("$$STOP"):
        STOPPINGING = True
        await message.channel.send(f"Stopped pinging :) you can now rest in peace :thumbsup:")
        await DebugChannel.send(f"Stopped pinging :) :thumbsdown:")
    if message.content.startswith("$$PING"):
      num = 0
      try:
        num = int(message.content[6:])
      except Exception as err:
        print(f"Error ocurred while parsing num of $$PINGs")
        await DebugChannel.send(f"Error ocurred while parsing num of $$PINGs")
        raise err
      if num <= 0:
        await DebugChannel.send("Number of pings left is <= 0, stopping !")
        return
      await PING()
      await message.channel.send(f"$$PING{num-1}")
    if message.author == client.user:
        return
    print(f"Message received: {message.content}")
    if message.content.startswith("$hello"):
        await message.channel.send("Hello!")
    if message.content.startswith("$member"):
        DebugChannel = message.channel
        for member in client.get_all_members():
            await message.channel.send(f"Member id: {member.id}, member name: {member.name}")
            if member.id == TargetID:
                await message.channel.send(f"Found my pinging target!! TIME TO PING! {member.id} {member.name}")
                TargetMember = member
    if message.content.startswith("$PingMember"):
        await PING()


async def PING(delay: int = 5):
    if STOPPINGING:
        await DebugChannel.send("Pings have been disabled :)")
        time.sleep(delay)
        return "STOPPINGING flag on :)"
    try:
        TargetChannel = await TargetMember.create_dm()
        await TargetChannel.send(
            "Hey bro man if you are seeing this than yay it worked!\nRemember to pay me **$10** :) :thumbsup:\nIf you want to stop this ping, type '$$STOP' :) or '$$START' to enable")
        time.sleep(delay)
    except Exception as err:
        await DebugChannel.send(f"Could not ping person :( :( {err}")
        raise err  # Ensures that things like keyboard interrupt are handled :)

client.run(TOKEN)



"""Includes classes and functions related to errors (and error handling)

Note: Basically every other module in this package (Smg88) depends on this one :) no pressure self
"""


class ErrorHandle():
    """Class that handles errors the user can identify / recover from
    """

    handleName = "Generic Error Handle: "

    def __init__(cls, msg, *msgs, **extra):
        msg = str(msg)
        _msgs = [str(m) for m in msgs]
        _msgs = '\n'.join(_msgs)
        cls.msg = f"{cls.handleName}\n{msg}\n{_msgs}"
        cls.extras = extra

    def __str__(self):
        return f"{self.msg}"


class UserErrorHandle(ErrorHandle):
    handleName = "User Error Handle: "


class ProgrammerErrorHandle(ErrorHandle):
    handleName = "Programmer Error Handle: "


class Error(Exception):
    """Exception class that is root of all custom exceptions
    """

    def __init__(self, msg, *msgs, errorHandle: ErrorHandle = ..., **extra):
        msg = str(msg)
        _msgs = [str(m) for m in msgs]
        _msgs = '\n'.join(_msgs)
        self.msg = f"Message: \n{msg}\n{_msgs}"
        self.extras = extra
        self.errorHandle = errorHandle

    def __str__(self):
        return f"{self.msg}\n" + "\n".join(f"{k}: {v}" for k, v in self.extras.items()) + f"\n{str(self.errorHandle)}"


class SafeCatchAll(Error, Exception):
    """Use instead of "catch Exception:" to catch all safe-to-catch exceptions

    This is to allow for a custom error deriving from errors.Error to indicate a system shutdown and not have to change any code
    """
    ...


class ProgrammerError(Error):
    """Exception class that is root of all internal / implementation / programmer related errors
    """
    ...


class InappropriateRequest(ProgrammerError):
    """Exception class that is raised when a request on some programming object / function / property is inappropriate
    """
    ...


class SimpleUserError(Error):
  """Error class used to refer to errors that are not programmer errors

  Args:
      Error (str / strings): Represents the error message and attached error handles
  """
  ...



"""This module contains the security-related functions for the Smg88 package.

Most of the objects in this module have their metaclass being of this class, namely 'Security'
This means that each binding (such as methods or properties) each have a security level from 0 to 69, where 0 is the most secure and 69 is the least secure.
"""


class Security(type):
  """This class is designed to be the metaclass of all objects following the Security design pattern.

  This makes the class only accessable to allowed security levels, with any other level than 69 needing confirmation
  """
  # TODO DESIGN then implement the security features proposed in the above docstrings
  ...



from ast import excepthandler
import json
import os
import shutil
from glob import glob

Download_Folder: str = "YAY SEE!"
Destination_Folder: str = "C:/Users/verys/Downloads/TestDestination/"

Tax_Return_Check_String: str = "signal"
Client_Contact_Check_String: str = "Minecraft"

# Get settings from basic cache file json format
__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))
_config_location = os.path.join(__location__, "settings.json")
action = input(
    "PRESS ENTER to continue\nPress 'r' to reset the settings file\nClose this windows to exit\n")
if action == "r":
  try:
    os.remove(_config_location)
    print("Great! Removed settings file, will now create a new one ...")
  except Exception as err:
    print(f"No settings file found, {err=}")
try:
  with open(_config_location, "x"):
    print("First time detected: Creating settings file :)")
except:
  print("Settings file detected :)")
with open(_config_location, "r") as config:
  settings = config.read()
  if settings == "":
    settings = "{}"
  settings = json.loads(settings)
  print("Reading config file ...")
  try:
    #print(f"{settings['Download_Folder']=}")
    #print(f"{type(settings['Download_Folder'])=}")
    if "Download_Folder" not in settings:
      settings["Download_Folder"] = input(
          "No download folder detected (copy paste below pls): ") + "/*"
      settings["Download_Folder"]
    Download_Folder = settings["Download_Folder"]
  except:
    ...
  try:
    if "Destination_Folder" not in settings:
      settings["Destination_Folder"] = input(
          "No destination folder detected (copy paste below pls): ")
    Destination_Folder = settings["Destination_Folder"]
  except:
    ...
  try:
    if "Tax_Return_Check_String" not in settings:
      settings["Tax_Return_Check_String"] = input(
          "No tax return check string detected (copy paste below pls): ")
    Tax_Return_Check_String = settings["Tax_Return_Check_String"]
  except:
    ...
  try:
    if "Client_Contact_Check_String" not in settings:
      settings["Client_Contact_Check_String"] = input(
          "No client contact check string detected (copy paste below pls): ")
    Client_Contact_Check_String = settings["Client_Contact_Check_String"]
  except:
    ...
settings = f'{{\n"Download_Folder":"{Download_Folder!r}",\n"Destination_Folder":"{Destination_Folder!r}",\n"Tax_Return_Check_String":"{Tax_Return_Check_String!r}",\n"Client_Contact_Check_String":"{Client_Contact_Check_String!r}"\n}}'
try:
  with open(_config_location, "w") as config:
    config.write(settings)
    print("Successfully updated settings :)")
except:
  print("Unable to update settings :(")

Ignore_Next_Tax_Return_Check: bool = False
Ignore_Next_Client_Contact_Check: bool = False

try:
  Files = glob(Download_Folder)
except Exception as err:
  print("Attempting to open download folder, maybe incorrect path given?")
  raise err
#print(f"Your download files: \n{Files}")
Files.sort(key=os.path.getctime)
try:
  Files.reverse()
except Exception as err:
  print("Attempting to reverse download files, maybe path contains no files?")
  raise err
if not os.path.isdir(Destination_Folder):
  print("Creating destination folder (given one does not exist) ...")
  try:
      os.mkdir(Destination_Folder)
  except OSError as err:
      print("Creation of the directory {Destination_Folder} failed")
      raise Exception(
          "Cannot find / create directory given to copy to\nTry to reset the settings file by chosing 'r' next run")
print("Destination Folder is valid (or created :)")
for file in Files:
  if Tax_Return_Check_String in file and not Ignore_Next_Tax_Return_Check:
    input(
        f"Moving this file: \n  {file}\n to folder \n  {Destination_Folder}\nClose this window to exit")
    try:
      shutil.move(file, Destination_Folder)
    except Exception as err:
      print(
          f"Unable to move file {file} to {Destination_Folder} for some reason (maybe destination not possible to save files in?)")
      raise err
    Ignore_Next_Tax_Return_Check = True
  if Client_Contact_Check_String in file and not Ignore_Next_Client_Contact_Check:
    input(
        f"Moving this file: \n  {file}\n to folder \n  {Destination_Folder}\nClose this window to exit")
    try:
      shutil.move(file, Destination_Folder)
    except Exception as err:
      print(
          f"Unable to move file {file} to {Destination_Folder} for some reason (maybe destination not possible to save files in?)")
      raise err
    Ignore_Next_Client_Contact_Check = True
if not Ignore_Next_Client_Contact_Check:
  print("OOPs, no Client Contact file found!")
if not Ignore_Next_Tax_Return_Check:
  print("OOPs, no Tax Return file found!")
